import cached_conv as cc
import gin
import pytest
import torch

from msprior.attention import Prior
from msprior.scripted import ScriptedPrior

torch.set_grad_enabled(False)
torch.random.manual_seed(42)
import msprior


def load_configuration(path, overrides=[]):

    def with_config_decorator(fun):

        def wrapper(*args, **kwargs):
            gin.clear_config()
            gin.parse_config_file(path)
            for override in overrides:
                gin.parse_config(override)
            out = fun(*args, **kwargs)
            return out

        return wrapper

    return with_config_decorator


@load_configuration(
    path="decoder_only.gin",
    overrides=["MODEL_DIM = 16", "HEAD_DIM = 2", "NUM_LAYERS = 2"],
)
def test_decoder_only():
    cc.use_cached_conv(False)
    regular = Prior().eval()

    cc.use_cached_conv(True)
    streaming = Prior().eval()

    for p1, p2 in zip(regular.parameters(), streaming.parameters()):
        p2.data.copy_(p1.data)

    input_params = gin.get_bindings("attention.MultivariateEmbedding")
    inputs = torch.randint(
        0,
        input_params["num_tokens"],
        (1, 128, input_params["num_quantizers"]),
    )

    outputs_regular = regular(inputs, targets=inputs)

    num_chunks = 4
    outputs_streaming = []
    for inputs in inputs.chunk(num_chunks, 1):
        outputs_streaming.append(streaming(inputs, targets=inputs))
    outputs_streaming = torch.cat(outputs_streaming, 1)

    assert torch.allclose(outputs_regular, outputs_streaming, 1e-4, 1e-4)


@load_configuration(
    path="decoder_only.gin",
    overrides=["MODEL_DIM = 16", "HEAD_DIM = 2", "NUM_LAYERS = 2"],
)
def test_decoder_only_nn_tilde():
    cc.use_cached_conv(False)
    regular = Prior().eval()

    cc.use_cached_conv(True)
    streaming = ScriptedPrior().eval()

    for p1, p2 in zip(regular.parameters(), streaming.parameters()):
        p2.data.copy_(p1.data)

    input_params = gin.get_bindings("attention.MultivariateEmbedding")
    inputs = torch.randint(
        0,
        input_params["num_tokens"],
        (1, 1, input_params["num_quantizers"]),
    )

    streaming.set_listen(False)
    streaming.set_reset(True)
    streaming.set_temperature(0)

    # AUTOREGRESSIVE SAMPLE USING REGULAR MODEL
    regular_samples = [inputs]
    for _ in range(4):
        regular_samples.append(
            regular.sample(
                torch.cat(regular_samples, 1),
                temperature=0,
            )[0][:, -1:])
    regular_samples = torch.cat(regular_samples, 1)

    # AUTOREGRESSIVE SAMPLE USING NN_TILDE
    streaming.previous_step[:1].copy_(inputs)
    out = streaming(torch.zeros(1, 16, 4))[:, :-1]
    streaming_samples = torch.cat([inputs, out.permute(0, 2, 1)], 1).long()

    assert torch.allclose(regular_samples, streaming_samples)



from typing import Callable, Dict, Optional, Tuple

import gin
import lmdb
import torch
from torch.utils.data import Dataset, Subset
from udls import AudioExample

TensorDict = Dict[str, torch.Tensor]


def ordered_dataset_split(dataset: Dataset,
                          sizes: Tuple[int, ...]) -> Tuple[Dataset]:
    assert sum(sizes) == len(dataset)
    datasets = []
    start_idx = 0
    for size in sizes:
        datasets.append(Subset(dataset, range(start_idx, start_idx + size)))
        start_idx += size
    return tuple(datasets)


@gin.configurable
class SequenceDataset(Dataset):

    def __init__(
            self,
            db_path: str,
            task_fn: Optional[Callable[[TensorDict],
                                       TensorDict]] = None) -> None:
        super().__init__()
        self._env = None
        self._keys = None
        self._db_path = db_path
        self.task_fn = task_fn

    def __len__(self):
        return len(self.keys)

    def __getitem__(self, index):
        with self.env.begin() as txn:
            ae = AudioExample(txn.get(self.keys[index]))
        ae = ae.as_dict()
        if self.task_fn is not None:
            ae = self.task_fn(ae)
        return ae

    @property
    def env(self):
        if self._env is None:
            self._env = lmdb.open(
                self._db_path,
                lock=False,
                readahead=False,
            )
        return self._env

    @property
    def keys(self):
        if self._keys is None:
            with self.env.begin(write=False) as txn:
                self._keys = list(txn.cursor().iternext(values=False))
        return self._keys



# Adapted from https://github.com/BlinkDL/RWKV-LM

import logging
import math
import os
from typing import Callable, Optional

import cached_conv as cc
import torch
import torch.nn as nn
from torch.utils.cpp_extension import load

T_MAX = 256

wkv_cuda = None


class WKV(torch.autograd.Function):

    @staticmethod
    def forward(ctx, B, T, C, w, u, k, v):
        ctx.B = B
        ctx.T = T
        ctx.C = C
        assert T <= T_MAX
        assert B * C % min(C, 1024) == 0
        w = -torch.exp(w.contiguous())
        u = u.contiguous()
        k = k.contiguous()
        v = v.contiguous()
        ctx.save_for_backward(w, u, k, v)
        y = torch.empty((B, T, C),
                        device='cuda',
                        memory_format=torch.contiguous_format)
        wkv_cuda.forward(B, T, C, w, u, k, v, y)
        return y

    @staticmethod
    def backward(ctx, gy):
        B = ctx.B
        T = ctx.T
        C = ctx.C
        assert T <= T_MAX
        assert B * C % min(C, 1024) == 0
        w, u, k, v = ctx.saved_tensors
        gw = torch.zeros((B, C), device='cuda').contiguous()
        gu = torch.zeros((B, C), device='cuda').contiguous()
        gk = torch.zeros((B, T, C), device='cuda').contiguous()
        gv = torch.zeros((B, T, C), device='cuda').contiguous()
        wkv_cuda.backward(B, T, C, w, u, k, v, gy.contiguous(), gw, gu, gk, gv)
        gw = torch.sum(gw, dim=0)
        gu = torch.sum(gu, dim=0)
        return (None, None, None, gw, gu, gk, gv)


def WKV_Attention(B, T, C, w, u, k, v):
    global wkv_cuda
    if wkv_cuda is None:
        logging.info("Building custom CUDA kernels")
        wkv_cuda = load(
            name=f"wkv_{T_MAX}",
            sources=[
                os.path.join(os.path.dirname(__file__), "cuda/wkv_op.cpp"),
                os.path.join(os.path.dirname(__file__), "cuda/wkv_cuda.cu"),
            ],
            verbose=False,
            extra_cuda_cflags=[
                "-t 4", "-std=c++17", "-res-usage", "--maxrregcount 60",
                "--use_fast_math", "-O3", "-Xptxas -O3",
                "--extra-device-vectorization", f"-DTmax={T_MAX}"
            ],
        )
    return WKV.apply(B, T, C, w, u, k, v)


class RWKV_TimeMix(nn.Module):

    def __init__(self, layer_id: int, ctx_len: int, dim: int, num_layers: int):
        super().__init__()
        self.layer_id = layer_id
        self.ctx_len = ctx_len
        self.dim = dim

        with torch.no_grad():  # fancy init
            ratio_0_to_1 = (layer_id / (num_layers - 1))  # 0 to 1
            ratio_1_to_almost0 = (1.0 - (layer_id / num_layers))  # 1 to ~0

            # time_decay init
            decay_speed = torch.ones(dim)
            for h in range(dim):
                decay_speed[h] = -5 + 8 * (h / (dim - 1))**(0.7 +
                                                            1.3 * ratio_0_to_1)
            self.time_decay = nn.Parameter(decay_speed)

            # time_first init
            time_first = (torch.tensor([(i + 1) % 3 - 1
                                        for i in range(dim)]) * 0.5)
            self.time_first = nn.Parameter(
                torch.ones(dim) * math.log(0.3) + time_first)

            # time_mix init
            x = torch.ones(1, 1, dim)
            for i in range(dim):
                x[0, 0, i] = i / dim
            self.time_mix_k = nn.Parameter(torch.pow(x, ratio_1_to_almost0))
            self.time_mix_v = nn.Parameter(
                torch.pow(x, ratio_1_to_almost0) + 0.3 * ratio_0_to_1)
            self.time_mix_r = nn.Parameter(
                torch.pow(x, 0.5 * ratio_1_to_almost0))

        self.time_shift = nn.ZeroPad2d((0, 0, 1, -1))

        self.key = nn.Linear(dim, dim, bias=False)
        self.value = nn.Linear(dim, dim, bias=False)
        self.receptance = nn.Linear(dim, dim, bias=False)

        self.output = nn.Linear(dim, dim, bias=False)

        self.key.scale_init = 0
        self.receptance.scale_init = 0
        self.output.scale_init = 0

    def forward(self, x):
        B, T, C = x.size()  # x = (Batch,Time,Channel)

        # Mix x with the previous timestep to produce xk, xv, xr
        xx = self.time_shift(x)
        xk = x * self.time_mix_k + xx * (1 - self.time_mix_k)
        xv = x * self.time_mix_v + xx * (1 - self.time_mix_v)
        xr = x * self.time_mix_r + xx * (1 - self.time_mix_r)

        # Use xk, xv, xr to produce k, v, r
        k = self.key(xk)
        v = self.value(xv)
        r = self.receptance(xr)
        sr = torch.sigmoid(r)
        rwkv = sr * WKV_Attention(
            B,
            T,
            C,
            self.time_decay,
            self.time_first,
            k,
            v,
        )
        rwkv = self.output(rwkv)
        return rwkv


class RWKV_ChannelMix(nn.Module):

    def __init__(self, layer_id: int, num_layers: int, dim: int):
        super().__init__()
        self.layer_id = layer_id

        self.time_shift = nn.ZeroPad2d((0, 0, 1, -1))

        with torch.no_grad():  # fancy init of time_mix
            ratio_1_to_almost0 = (1.0 - (layer_id / num_layers))  # 1 to ~0

            x = torch.ones(1, 1, dim)
            for i in range(dim):
                x[0, 0, i] = i / dim

            self.time_mix_k = nn.Parameter(torch.pow(x, ratio_1_to_almost0))
            self.time_mix_r = nn.Parameter(torch.pow(x, ratio_1_to_almost0))

        hidden_sz = 4 * dim
        self.key = nn.Linear(dim, hidden_sz, bias=False)
        self.receptance = nn.Linear(dim, dim, bias=False)
        self.value = nn.Linear(hidden_sz, dim, bias=False)

        self.value.scale_init = 0
        self.receptance.scale_init = 0

    def forward(self, x):
        xx = self.time_shift(x)
        xk = x * self.time_mix_k + xx * (1 - self.time_mix_k)
        xr = x * self.time_mix_r + xx * (1 - self.time_mix_r)

        k = self.key(xk)
        k = torch.square(torch.relu(k))
        kv = self.value(k)

        rkv = torch.sigmoid(self.receptance(xr)) * kv
        return rkv


class IncrementalTimeMix(nn.Module):

    def __init__(self, layer_id: int, ctx_len: int, dim: int, num_layers: int):
        super().__init__()
        self.layer_id = layer_id
        self.ctx_len = ctx_len
        self.dim = dim

        self.time_decay = nn.Parameter(torch.empty(dim))
        self.time_first = nn.Parameter(torch.empty(dim))
        self.time_mix_k = nn.Parameter(torch.empty(dim))
        self.time_mix_v = nn.Parameter(torch.empty(dim))
        self.time_mix_r = nn.Parameter(torch.empty(dim))

        self.key = nn.Linear(dim, dim, bias=False)
        self.value = nn.Linear(dim, dim, bias=False)
        self.receptance = nn.Linear(dim, dim, bias=False)

        self.output = nn.Linear(dim, dim, bias=False)

        self.register_buffer("last_x", torch.zeros(cc.MAX_BATCH_SIZE, 1, dim))
        self.register_buffer("last_num", torch.zeros(cc.MAX_BATCH_SIZE, 1,
                                                     dim))
        self.register_buffer("last_den", torch.zeros(cc.MAX_BATCH_SIZE, 1,
                                                     dim))

    def interpolate(self, x: torch.Tensor,
                    factor: torch.Tensor) -> torch.Tensor:
        x = x * factor + self.last_x[:x.shape[0]] * (1 - factor)
        return x

    def update_cache(self, x: torch.Tensor, num: torch.Tensor,
                     den: torch.Tensor) -> None:
        batch_size = x.shape[0]

        self.last_x[:batch_size].copy_(x)
        self.last_num[:batch_size].copy_(num)
        self.last_den[:batch_size].copy_(den)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        batch_size = x.shape[0]
        k = self.key(self.interpolate(x, self.time_mix_k))
        v = self.value(self.interpolate(x, self.time_mix_v))
        r = self.receptance(self.interpolate(x, self.time_mix_r))

        exp_k = torch.exp(self.time_first + k)
        wkv = (self.last_num[:batch_size] +
               exp_k * v) / (self.last_den[:batch_size] + exp_k)
        rwkv = torch.sigmoid(r) * wkv

        decay = torch.exp(-torch.exp(self.time_decay))
        exp_k = torch.exp(k)
        num = decay * self.last_num[:batch_size] + exp_k * v
        den = decay * self.last_den[:batch_size] + exp_k

        out = self.output(rwkv)

        self.update_cache(x, num, den)
        return out


class IncrementalChannelMix(nn.Module):

    def __init__(self, layer_id: int, num_layers: int, dim: int):
        super().__init__()
        self.layer_id = layer_id

        self.time_mix_k = nn.Parameter(torch.empty(dim))
        self.time_mix_r = nn.Parameter(torch.empty(dim))

        hidden_sz = 4 * dim
        self.key = nn.Linear(dim, hidden_sz, bias=False)
        self.receptance = nn.Linear(dim, dim, bias=False)
        self.value = nn.Linear(hidden_sz, dim, bias=False)

        self.register_buffer("last_x", torch.zeros(cc.MAX_BATCH_SIZE, 1, dim))

    def interpolate(self, x: torch.Tensor,
                    factor: torch.Tensor) -> torch.Tensor:
        x = x * factor + self.last_x[:x.shape[0]] * (1 - factor)
        return x

    def update_cache(self, x: torch.Tensor) -> None:
        batch_size = x.shape[0]

        self.last_x[:batch_size].copy_(x)

    def forward(self, x: torch.Tensor):
        k = self.key(self.interpolate(x, self.time_mix_k))
        r = self.receptance(self.interpolate(x, self.time_mix_r))

        vk = self.value(torch.relu(k).pow(2))

        out = torch.sigmoid(r) * vk

        self.update_cache(x)
        return out


class Block(nn.Module):

    def __init__(self, layer_id: int, dim: int,
                 time_mixer: Callable[[int], RWKV_TimeMix],
                 channel_mixer: Callable[[int], RWKV_ChannelMix]) -> None:
        super().__init__()
        self.t_mixer = nn.Sequential(nn.LayerNorm(dim), time_mixer(layer_id))
        self.c_mixer = nn.Sequential(nn.LayerNorm(dim),
                                     channel_mixer(layer_id))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.t_mixer(x) + x
        x = self.c_mixer(x) + x
        return x


class Film(nn.Module):

    def __init__(self, in_dim: int, feature_dim: int) -> None:
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(in_dim, feature_dim),
            nn.SiLU(),
            nn.Linear(feature_dim, 2 * feature_dim),
        )

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        mean, scale = self.net(y).chunk(2, -1)
        return x * scale + mean


class RWKV(nn.Module):

    def __init__(self,
                 block: Callable[[int], Block],
                 num_layers: int,
                 dim: int,
                 film: Optional[Callable[[], Film]] = None) -> None:
        super().__init__()
        self.norm = nn.LayerNorm(dim)
        self.blocks = nn.ModuleList([block(i) for i in range(num_layers)])
        self.films = None
        if film is not None:
            self.films = nn.ModuleList([film() for _ in range(num_layers)])

    def forward(self,
                x: torch.Tensor,
                y: Optional[torch.Tensor] = None) -> torch.Tensor:
        if cc.USE_BUFFER_CONV and x.shape[1] != 1:
            out = []
            if self.films is not None:
                assert y is not None
                for _x, _y in zip(
                        x.chunk(x.shape[1], 1),
                        y.chunk(y.shape[1], 1),
                ):
                    _x = self.norm(_x)
                    for block, film in zip(self.blocks, self.films):
                        _x = block(_x)
                        _x = film(_x, _y)
                    out.append(_x)
            else:
                for _x in x.chunk(x.shape[1], 1):
                    _x = self.norm(_x)
                    for block in self.blocks:
                        _x = block(_x)
                    out.append(_x)
            x = torch.cat(out, 1)
        else:
            x = self.norm(x)
            if self.films is not None:
                assert y is not None
                for block, film in zip(self.blocks, self.films):
                    x = block(x)
                    x = film(x, y)
            else:
                for block in self.blocks:
                    x = block(x)
        return x


def TimeMixer(layer_id: int, ctx_len: int, dim: int, num_layers: int):
    if cc.USE_BUFFER_CONV:
        return IncrementalTimeMix(layer_id, ctx_len, dim, num_layers)
    else:
        return RWKV_TimeMix(layer_id, ctx_len, dim, num_layers)


def ChannelMixer(layer_id: int, num_layers: int, dim: int):
    if cc.USE_BUFFER_CONV:
        return IncrementalChannelMix(layer_id, num_layers, dim)
    else:
        return RWKV_ChannelMix(layer_id, num_layers, dim)



import logging
import os

import gin
import pytorch_lightning as pl
import torch
from absl import app, flags
from torch.utils import data

import msprior.utils
from msprior.attention import Prior
from msprior.dataset import SequenceDataset

FLAGS = flags.FLAGS
flags.DEFINE_multi_string("config",
                          default="decoder_only",
                          help="config to parse.")
flags.DEFINE_string("db_path",
                    default=None,
                    required=True,
                    help="path to dataset.")
flags.DEFINE_integer("val_size",
                     default=8192,
                     help="size of validation dataset.")
flags.DEFINE_integer("batch_size", default=64, help="batch size.")
flags.DEFINE_string("name", default=None, required=True, help="train name.")
flags.DEFINE_integer("gpu", default=0, help="gpu index.")
flags.DEFINE_integer("workers",
                     default=0,
                     help="num workers during data loading.")
flags.DEFINE_string("pretrained_embedding",
                    default=None,
                    help="use pretrained embeddings from rave.")
flags.DEFINE_multi_string("override",
                          default=[],
                          help="additional gin bindings.")
flags.DEFINE_string("ckpt",
                    default=None,
                    help="checkpoint to resume training from.")
flags.DEFINE_float('ema',
                   default=None,
                   help='Exponential weight averaging factor (optional)')
flags.DEFINE_integer("val_every",
                     default=1000,
                     help="validate training every n step.")


def add_ext(config: str):
    if config[-4:] != ".gin":
        config += ".gin"
    return config


def main(argv):
    logging.info("parsing configuration")
    configs = list(map(add_ext, FLAGS.config))

    overrides = FLAGS.override
    if FLAGS.pretrained_embedding is not None:
        overrides.append(f"PRETRAINED_RAVE='{FLAGS.pretrained_embedding}'")

    gin.parse_config_files_and_bindings(
        configs,
        overrides,
    )

    logging.info("loading dataset")
    dataset = SequenceDataset(db_path=FLAGS.db_path)

    if FLAGS.val_size > len(dataset):
        logging.warn(
            r"Dataset too small, using 5% of the train set as the val set")
        FLAGS.val_size = len(dataset) // 20

    train, val = data.random_split(
        dataset,
        (len(dataset) - FLAGS.val_size, FLAGS.val_size),
        generator=torch.Generator().manual_seed(42),
    )
    if not any(map(lambda x: "flattened" in x, FLAGS.config)):
        logging.info("quantizer number retrieval")
        with gin.unlock_config():
            gin.parse_config(
                f"NUM_QUANTIZERS={train[0]['decoder_inputs'].shape[-1]}")

    logging.info("building model")
    model = Prior()

    train_loader = data.DataLoader(
        train,
        batch_size=FLAGS.batch_size,
        shuffle=True,
        drop_last=True,
        num_workers=FLAGS.workers,
    )
    val_loader = data.DataLoader(
        val,
        batch_size=FLAGS.batch_size,
        shuffle=False,
        drop_last=False,
        num_workers=FLAGS.workers,
    )

    os.makedirs(os.path.join("runs", FLAGS.name), exist_ok=True)
    with open(os.path.join("runs", FLAGS.name, "config.gin"),
              "w") as config_out:
        config_out.write(gin.config_str())

    val_check = {}
    if len(train_loader) >= FLAGS.val_every:
        val_check["val_check_interval"] = FLAGS.val_every
    else:
        nepoch = FLAGS.val_every // len(train_loader)
        val_check["check_val_every_n_epoch"] = nepoch

    callbacks = [
        pl.callbacks.LearningRateMonitor(logging_interval='step'),
        pl.callbacks.ModelCheckpoint(monitor="val_cross_entropy",
                                     filename='best'),
        pl.callbacks.ModelCheckpoint(filename='last'),
        pl.callbacks.EarlyStopping(
            "val_cross_entropy",
            patience=20,
        )
    ]

    if FLAGS.ema is not None:
        callbacks.append(msprior.utils.EMA(FLAGS.ema))

    logging.info("creating trainer")
    trainer = pl.Trainer(
        logger=pl.loggers.TensorBoardLogger('runs', name=FLAGS.name),
        accelerator='gpu',
        devices=[FLAGS.gpu],
        callbacks=callbacks,
        log_every_n_steps=10,
        **val_check,
    )

    torch.backends.cudnn.benchmark = True
    torch.set_float32_matmul_precision('high')

    logging.info("launch training")
    trainer.fit(
        model,
        train_loader,
        val_loader,
        ckpt_path=FLAGS.ckpt,
    )


if __name__ == "__main__":
    app.run(main)



import cached_conv as cc
import nn_tilde
import torch
from absl import flags


class CombinedModule(nn_tilde.Module):

    def __init__(self, prior, rave) -> None:
        super().__init__()
        self.prior_model = torch.jit.load(prior).eval()
        self.rave_model = torch.jit.load(rave).eval()

        self.register_attribute("temperature",
                                self.prior_model.get_temperature())
        self.register_attribute("listen", self.prior_model.get_listen())
        self.register_attribute("learn_context",
                                self.prior_model.get_learn_context())
        self.register_attribute("reset", self.prior_model.get_reset())

        self.register_method(
            "encode",
            *self.rave_model.encode_params,
            self.rave_model.encode_input_labels,
            self.rave_model.encode_output_labels,
        )
        self.register_method(
            "decode",
            *self.rave_model.decode_params,
            self.rave_model.decode_input_labels,
            self.rave_model.decode_output_labels,
        )
        self.register_method(
            "forward",
            *self.rave_model.forward_params,
            self.rave_model.forward_input_labels,
            self.rave_model.forward_output_labels,
        )
        self.register_method(
            "prior",
            *self.prior_model.forward_params,
            self.prior_model.forward_input_labels,
            self.prior_model.forward_output_labels,
        )

    @torch.jit.export
    def encode(self, x):
        return self.rave_model.encode(x)

    @torch.jit.export
    def decode(self, x):
        return self.rave_model.decode(x)

    @torch.jit.export
    def forward(self, x):
        return self.rave_model.forward(x)

    @torch.jit.export
    def prior(self, x):
        return self.prior_model.forward(x)

    @torch.jit.export
    def get_temperature(self) -> float:
        return self.prior_model.get_temperature()

    @torch.jit.export
    def set_temperature(self, temperature: float) -> int:
        return self.prior_model.set_temperature(temperature)

    @torch.jit.export
    def get_listen(self) -> bool:
        return self.prior_model.get_listen()

    @torch.jit.export
    def set_listen(self, listen: bool) -> int:
        return self.prior_model.set_listen(listen)

    @torch.jit.export
    def get_learn_context(self) -> bool:
        return self.prior_model.get_learn_context()

    @torch.jit.export
    def set_learn_context(self, learn_context: bool) -> int:
        return self.prior_model.set_learn_context(learn_context)

    @torch.jit.export
    def get_reset(self) -> bool:
        return self.prior_model.get_reset()

    @torch.jit.export
    def set_reset(self, reset: bool) -> int:
        return self.prior_model.set_reset(reset)


def main(argv):
    cc.MAX_BATCH_SIZE = FLAGS.batch_size
    CombinedModule(FLAGS.prior, FLAGS.rave).export_to_ts(FLAGS.name + ".ts")


FLAGS = flags.FLAGS
flags.DEFINE_string(
    'prior',
    default=None,
    required=True,
    help='Prior model',
)
flags.DEFINE_string(
    'rave',
    default=None,
    help='Streaming rave model',
)
flags.DEFINE_integer(
    'batch_size',
    default=1,
    help='Maximum batch size',
)
flags.DEFINE_string(
    'name',
    default=None,
    required=True,
    help="Combined model name",
)



from random import randint
from typing import Dict

import numpy as np
import torch

TensorDict = Dict[str, torch.Tensor]


def decoder_only_rave(inputs: TensorDict, seq_len: int, decoder_key: str,
                      quantizer_crop: int) -> TensorDict:
    start = randint(0, inputs[decoder_key].shape[0] - seq_len - 1)
    end = start + seq_len

    input_tokens = inputs[decoder_key][start:end, :quantizer_crop].copy()
    target_tokens = inputs[decoder_key][start + 1:end +
                                        1, :quantizer_crop].copy()

    return {
        "encoder_inputs": torch.Tensor([]),
        "decoder_inputs": input_tokens,
        "decoder_targets": target_tokens,
    }


def decoder_only_rave_flattened(inputs: TensorDict, seq_len: int,
                                decoder_key: str, num_tokens: int,
                                quantizer_crop: int) -> TensorDict:
    inputs = inputs[decoder_key][:, :quantizer_crop]
    inputs = inputs + np.arange(quantizer_crop) * num_tokens
    inputs = inputs.reshape(-1)
    start = randint(0, inputs.shape[0] - seq_len - 1)
    if start % quantizer_crop:
        start -= start % quantizer_crop
    end = start + seq_len

    input_tokens = inputs[start:end].copy()
    target_tokens = inputs[start + 1:end + 1].copy()

    return {
        "encoder_inputs": torch.Tensor([]),
        "decoder_inputs": input_tokens,
        "decoder_targets": target_tokens,
    }


def encoder_decoder_semantic_token(inputs: TensorDict, seq_len: int,
                                   encoder_key: str,
                                   decoder_key: str) -> TensorDict:
    start = randint(0, inputs[decoder_key].shape[0] - seq_len - 1)
    ratio = inputs[decoder_key].shape[0] // inputs[encoder_key].shape[0]
    start = start - (start % ratio)
    end = start + seq_len

    input_tokens = inputs[decoder_key][start:end].copy()
    target_tokens = inputs[decoder_key][start + 1:end + 1].copy()

    semantic_indices = inputs[encoder_key].copy()
    semantic_indices = semantic_indices[start // ratio:end // ratio]

    return {
        "encoder_inputs": semantic_indices,
        "decoder_inputs": input_tokens,
        "decoder_targets": target_tokens,
    }


def encoder_decoder_semantic_features(inputs: TensorDict, seq_len: int,
                                      encoder_key: str,
                                      decoder_key: str) -> TensorDict:
    start = randint(0, inputs[decoder_key].shape[0] - seq_len - 1)
    ratio = inputs[decoder_key].shape[0] // inputs[encoder_key].shape[0]
    start = start - (start % ratio)
    end = start + seq_len

    input_tokens = inputs[decoder_key][start:end].copy()
    target_tokens = inputs[decoder_key][start + 1:end + 1].copy()

    semantic_features = inputs[encoder_key].copy()
    semantic_features = semantic_features[start // ratio:end // ratio]

    return {
        "encoder_inputs": semantic_features,
        "decoder_inputs": input_tokens,
        "decoder_targets": target_tokens,
    }


def conditioned_rwkv(inputs: TensorDict, seq_len: int, main_key: str,
                     condition_key: str) -> TensorDict:
    main_seq = inputs[main_key]
    condition_seq = inputs[condition_key]
    ratio = main_seq.shape[0] // condition_seq.shape[0]

    if ratio < 1:
        raise ValueError(
            f"time ratio between main and condition should be >= 1, got {ratio}"
        )

    condition_seq = np.repeat(condition_seq, ratio, axis=0)

    start = randint(0, main_seq.shape[0] - seq_len - 1)
    end = start + seq_len

    return {
        "encoder_inputs": condition_seq[start + 1:end + 1],
        "decoder_inputs": main_seq[start:end],
        "decoder_targets": main_seq[start + 1:end + 1],
    }



import os

import gin

gin.add_config_file_search_path(os.path.dirname(__file__))
gin.add_config_file_search_path(
    os.path.join(os.path.dirname(__file__), 'configs'))

for subdir in ["arch", "condition", "type"]:
    gin.add_config_file_search_path(
        os.path.join(os.path.dirname(__file__), 'configs', subdir))



import os

import cached_conv as cc
import torch
from absl import flags

from msprior.scripted import ScriptedPrior

torch.set_grad_enabled(False)


def main(argv):
    cc.MAX_BATCH_SIZE = FLAGS.batch_size
    cc.use_cached_conv(True)

    model = ScriptedPrior(
        run=FLAGS.run,
        temporal_ratio=FLAGS.temporal_ratio,
        from_continuous=FLAGS.continuous,
        vae_path=FLAGS.vae_path,
        ema_weights=FLAGS.ema_weights,
    )
    model.apply_full_reset()
    model_name = os.path.basename(os.path.normpath(FLAGS.run)) + '.ts'
    export_path = os.path.join(FLAGS.run, model_name)
    model.export_to_ts(export_path)
    print(f'model exported to {export_path}')


FLAGS = flags.FLAGS
flags.DEFINE_string('run', default=None, required=True, help='Run to export')
flags.DEFINE_string('vae_path', default=None, help='Pretrained semantic VAE')
flags.DEFINE_integer('batch_size', default=1, help='Maximum batch size')
flags.DEFINE_bool('ema_weights',
                  default=False,
                  help='Use ema weights if available')

flags.DEFINE_integer(
    'temporal_ratio',
    default=1024,
    help='Sequence temporal ratio',
)
flags.DEFINE_bool(
    'continuous',
    default=False,
    help=
    'dequantize predictions if rave model is continuous rather than discrete',
)
